"""Awesome `data_manifestor` is a Python cli/package created with https://github.com/TezRomacH/python-package-template"""

import sys
from importlib import metadata as importlib_metadata

from .core import (
    compare_manifest_to_local,
    get_project_data_manifestor_template,
    path_to_manifest_template,
    update_project_data_manifestor_template,
)
from .exceptions import DataManifestorError, ManifestError
from .models import Template


def get_version() -> str:
    try:
        return importlib_metadata.version(__name__)
    except importlib_metadata.PackageNotFoundError:  # pragma: no cover
        return "unknown"


version: str = get_version()

__all__ = [
    "version",
    "compare_manifest_to_local",
    "get_project_data_manifestor_template",
    "update_project_data_manifestor_template",
    "path_to_manifest_template",
    "Template",
    "DataManifestorError",
    "ManifestError",
]



from typing import Any, Dict, List, Tuple, Union

from dataclasses import dataclass, field

from .schemas import Template as TemplateSchema


@dataclass
class Template:

    name: str
    path_prefix: str = "{}-{}-{}"
    path_patterns: list[str] = field(default_factory=list)


@dataclass
class Manifest:

    name: str
    template: Template
    args: tuple[str, ...] = field(default_factory=tuple)
    path_patterns: list[str] = field(default_factory=list)


@dataclass
class LocalComparisonResult:

    manifest: Manifest
    local_dir: str
    resolved_paths: list[tuple[str, list[str]]] = field(default_factory=list)
    found: list[str] = field(default_factory=list)
    missing: list[str] = field(default_factory=list)


template_schema = TemplateSchema()


def dict_to_template(data: dict[str, Any]) -> Any:
    loaded = template_schema.load(data)  # validation
    return Template(**loaded)


def alt_dict_to_template(data: dict[str, Any]) -> Template:
    """
    Note
    ----
    alternative template syntax from Ben
    """
    path_patterns = []
    for key, value in data["files"].items():
        filename = value.get("filename")
        directory_name = value.get("directory_name")
        if filename:
            path_patterns.append(filename.lstrip("%"))
        elif directory_name:
            path_patterns.append(directory_name.lstrip("%"))
        else:
            raise Exception("Invalid format for: %s" % value)

    return Template(
        name=data["name"],
        path_prefix="{}-{}-{}",
        path_patterns=path_patterns,
    )


def is_alt_format(data: dict[str, Any]) -> bool:
    return data.get("files") is not None


def dump_Template(template: Template) -> Any:
    return template_schema.dump(template)



from typing import Any, Dict

from marshmallow import Schema, fields


class Template(Schema):

    name = fields.String()
    path_prefix = fields.String(dump_default="{}-{}-{}")
    path_patterns = fields.List(fields.String())



class DataManifestorError(Exception):

    """Base exception type for all data_manifestor errors."""


class PathError(DataManifestorError):

    """An issue related to finding a path."""


class ManifestError(DataManifestorError):

    """An issue related to parsing the manifest."""



from typing import Any, List, Tuple

import os
from glob import glob


def resolve_paths(
    parent_dir: str,
    path_pattern: str,
) -> tuple[str, list[str]]:
    search_pattern = os.path.join(parent_dir, path_pattern)
    if os.path.isdir(search_pattern):
        search_pattern = os.path.join(
            search_pattern, "*"
        )  # so we can glob all children
    return (
        search_pattern,
        glob(search_pattern),
    )


def generate_path_pattern(
    path_prefix: str,
    path_pattern: str,
    *args: Any,
) -> str:
    full_pattern = path_prefix + path_pattern

    try:
        return full_pattern.format(*args)
    except IndexError:
        raise Exception(
            "Incompatible path_pattern and args combination. path_pattern=%s, args=%s",
            full_pattern,
            args,
        )



from typing import Any, Dict, Optional, Tuple

import logging
import os
from glob import glob

from np_lims_tk.core import get_project_asset, get_project_name, put_project_asset
from np_lims_tk.local_paths import path_to_lims_meta

from .exceptions import ManifestError, PathError
from .models import (
    LocalComparisonResult,
    Manifest,
    Template,
    alt_dict_to_template,
    dict_to_template,
    dump_Template,
    is_alt_format,
)
from .paths import generate_path_pattern, resolve_paths

logger = logging.getLogger(__name__)


def generate_manifest(template_dict: dict[str, Any], *args: str) -> Manifest:
    try:
        if is_alt_format(template_dict):
            template = alt_dict_to_template(template_dict)
        else:
            template = dict_to_template(template_dict)

        path_patterns = [
            generate_path_pattern(template.path_prefix, path_pattern, *args)
            for path_pattern in template.path_patterns
        ]
    except Exception as e:
        raise ManifestError("Error parsing manifest") from e

    return Manifest(
        name=template.name,
        template=template,
        args=args,
        path_patterns=path_patterns,
    )


def compare_manifest_to_local(
    local_dir: str,
    *args: str,
    template_dict: Optional[dict[str, Any]] = None,
) -> LocalComparisonResult:
    if not template_dict:
        logger.debug("Autogenerating manifest and args")
        template_dict, args = path_to_manifest_template(local_dir)
        if not template_dict:
            raise ManifestError("Couldnt infer template from path. path=%s" % local_dir)

    manifest = generate_manifest(template_dict, *args)
    resolved_paths = []
    found = []
    missing = []
    for path_pattern in manifest.path_patterns:
        pattern, paths = resolve_paths(local_dir, path_pattern)
        if len(paths) > 0:
            found.extend(paths)
        else:
            missing.append(pattern)
        resolved_paths.append(
            (
                pattern,
                paths,
            )
        )
    return LocalComparisonResult(
        local_dir=local_dir,
        manifest=manifest,
        resolved_paths=resolved_paths,
        found=found,
        missing=missing,
    )


def path_to_manifest_template(path: str) -> tuple[Any, tuple[Any, ...]]:
    """Infers associated manifest template from a file or directory path.

    Args
    ----
    path: File or directory path.

    Returns
    -------
    manifest template and necessary arguments associated with it

    Raises
    ------
    ManifestError
    - An unexpected error occurred
    """
    try:
        lims_meta = path_to_lims_meta(path)
        logger.debug("Resolved lims meta: %s" % lims_meta)
        lims_id = lims_meta._id
        subject_id = lims_meta.subject_id
        date_str = lims_meta.date_str
        project_name = get_project_name(lims_id=lims_id)
        logger.debug("Resolved project name: %s" % project_name)
        template = get_project_data_manifestor_template(project_name)
        return template, (
            lims_id,
            subject_id,
            date_str,
        )
    except Exception as e:
        raise ManifestError("Error parsing manifest") from e


data_manifestor_project_asset_name = "data_manifestor_template"


def get_project_data_manifestor_template(
    project_name: str = "default",
    client: Optional[Any] = None,  # for integration testing, remove later
) -> Any:
    return get_project_asset(
        project_name,
        data_manifestor_project_asset_name,
        client=client,
    )


def update_project_data_manifestor_template(
    project_name: str,
    template: Template,
    client: Optional[Any] = None,  # for integration testing, remove later
) -> None:
    put_project_asset(
        project_name,
        data_manifestor_project_asset_name,
        dump_Template(template),
        client=client,
    )
